#include "afflib.h"
#include "afflib_i.h"
#include "imager.h"
#include "aimage.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <err.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <term.h>
#include <assert.h>
#include <errno.h>

/*
 * imager.cpp:
 * The C++ imaging library.
 */

imager::imager()
{
    allow_regular = false;
    total_sectors_written = 0;
    total_sectors_read=0;
    total_bytes_read = 0;
    total_bytes_written = 0;
    total_blank_sectors = 0;
    
    callback_bytes_to_write = 0;
    callback_bytes_written = 0;

    imaging = false;
    imaging_failed = false;

    opt_zap = false;
    opt_append = false;
    opt_logAFF = false;
    logfile = 0;


    last_sector_read = 0;	// sector number
    bad_sectors_read = 0;
    af = 0;
    fout = 0;
    hash_invalid = false;

    memset(cmd_attach,0,sizeof(cmd_attach));
    memset(cmd_detach,0,sizeof(cmd_detach));

    scsi_bus = -1;
    scsi_tid = -1;
    scsi_lun = -1;
    scsi_pass = -1;
    ata_dev = -1;
    
    memset(device_model,0,sizeof(device_model));
    memset(serial_number,0,sizeof(serial_number));
    memset(firmware_revision,0,sizeof(firmware_revision));

    in = -1;
    sector_size = 0;
    total_sectors = 0;
    maxreadblocks = 0;

    af = 0;
    fout = 0;

    memset(fname_raw,0,sizeof(fname_raw));
    memset(fname_aff,0,sizeof(fname_aff));
    memset(infile,0,sizeof(infile));

    memset(final_md5,0,sizeof(final_md5));
    memset(final_sha1,0,sizeof(final_sha1));
    hash_invalid = false;

    last_sector_read = 0;
    last_sectors_read = 0;

    seek_on_output = false;
    retry_count = 0;

    buf = 0;
    bufsize = 512;			// good guess
    memset(blank_sector,0,sizeof(blank_sector));
    partial_sector_left  = 0;
    partial_sector_blank = false;

    bad_sectors_read = 0;
    errors_in_a_row = 0;
    error_recovery_phase = 0;
    last_direction = 0;

    /* error recovery */
}


void imager::write_data(unsigned char *buf,uint64 offset,int len)
{
    if(!hash_invalid){
	/* Update hash functions. */
	MD5_Update(&md5,buf,len);
	SHA1_Update(&sha,buf,len);
    }

    /* Count the number of blank sectors.
     */
    /* First, see if there is a partial blank sector that we are still processing... */
    int len_left = len;
    while(len_left>0 && partial_sector_left>0){
	if(buf[len-len_left] != 0){
	    partial_sector_blank = false; // it's no longer blank
	}
	len_left--;
	partial_sector_left--;
    }
    if(partial_sector_left==0){ // we reached the end of the partial sector
	if(partial_sector_blank==true){	// and the sector is blank!
	    total_blank_sectors++;
	}
    }
    /* Is it possible to look for full sectors? */
    while(len_left > sector_size){
	if(memcmp(buf+(len-len_left),blank_sector,sector_size)==0){
	    total_blank_sectors++;
	}
	len_left -= sector_size;
    }

    if(partial_sector_left==0 && len_left>0){ // some left, so we have a new partial sector
	partial_sector_left  = sector_size;
	partial_sector_blank = true;
    }

    /* If anything is left, do the partial sector */
    while(len_left>0){
	if(buf[len-len_left] != 0){
	    partial_sector_blank = false; // no longer blank
	}
	len_left--;
	partial_sector_left--;
    }


    /* Write it out and carry on... */
    if(af){
	af_seek(af,offset,SEEK_SET);
	if(af_write(af,buf,len)!=len){
	    perror("af_write");	// this is bad
	    af_close(af);	// try to gracefully recover
	    fprintf(stderr,"Imaging terminated because af_write failed.\n");
	    fprintf(stderr,"Correct error condition and re-run aimage. You should\n");
	    fprintf(stderr,"be able to carry on from where you left off.\n");
	    exit(1);
	}
    }
    if(fout){
	fseeko(fout,offset,SEEK_SET);
	if(fwrite(buf,1,len,fout)!=(unsigned)len){
	    perror("fwrite");	// this is also bad
	    fclose(fout);	// get what we can out to disk
	    fprintf(stderr,"Imaging terminated because fwrite() failed.\n");
	    fprintf(stderr,"Correct error condition and re-run aimage.\n");
	    fprintf(stderr,"You may be able to carry on from where you left off.\n");
	    exit(1);
	}
    }
    total_sectors_written += len/sector_size;
    total_bytes_written   += len;
}

void imager::status()
{
    my_refresh(this,0);			// just refresh; most status is done by AFF callback
}

/****************************************************************
 *** isleep(): An informative sleep...
 ****************************************************************/
void isleep(int s)
{
  printf("isleep %d\n",s);
  for(int i=0;i<s;i++){
    printf("\rSleeping for %d seconds; %d left...",s,s-i);
    fflush(stdout);
    sleep(1);
  }
  printf("\r%50s\r","");
  fflush(stdout);
}



/*
 * open_dev(char outdev[MAXPATHLEN],char *indev)
 * Try to open the friendly device name.
 * If successful, return the actual device in outdev and the FD.
 * 
 * If the device can be detected but not mounted (common with some
 * broken IDE drives), return fd==65536 (FD_IDENT). 
 * This says that we can't open it, but should ident it.
 */


int  imager::open_dev(const char *friendly_name)
{
    /****************************************************************
     *** Check for ata%d or ide%d
     ****************************************************************/

    ata_dev = -1;
    sscanf(friendly_name,"ata%d",&ata_dev);	// try to find ata0
    if(ata_dev==-1){
	sscanf(friendly_name,"ide%d",&ata_dev);	// try to find ide0
    }

    if(ata_dev != -1){			// if we found the device
	/* Create the attach and detach commands */

	char dev0[64];			// space for the first channel
	char dev1[64];			// space for the second channel
	char *dev[2] = {dev0,dev1};

	make_ata_attach_commands(cmd_attach,cmd_detach,dev0,dev1,ata_dev);
	system(cmd_detach);	// make sure we are detached first

	int i;
	for(i=0;i<10;i++){
	    int delay = i*3;
	    printf("\nOpening special ATA Bus #%d...\n",ata_dev);
	    if(i>0){
		printf("Attempt %d out of %d.\n",i+1,10);
	    }
	    printf("# %s\n",cmd_attach);
	    system(cmd_attach);
	    
	    /* See if we found the device */
	    for(int d=0;d<2;d++){
		if(access(dev[d],F_OK)==0){
		    if(access(dev[d],R_OK)){
			// don't have permission to read it.
			// this is bad
			err(1,dev[d]);
		    }
		    if(delay){
			printf("Waiting %d second%s for %s to spin up...\n",
			       delay,delay==1?"":"s",infile);  
			isleep(delay);
		    }	
		    strcpy(infile,dev[d]); // we will try this one
		    int fd = open(infile,O_RDONLY);
		    if(fd>0){
			/* The device was successfully opened. */
			return fd;		// got it!
		    }
		    perror(infile);		// give a warning
		}
	    }
	    printf("Detaching device and trying again...\n");
	    printf("# %s\n",cmd_detach);
	    system(cmd_detach);
	    isleep(delay);
	}
	/* Been through too many times. Did we get a device?
	 * If so, just ident it...
	 */
	if(infile[0]){
	    imaging_failed = true;
	    return FD_IDENT;
	}
    }

    /****************************************************************
     *** Check for scsi%d
     *** In our testing with FreeBSD, there is no advantage to repeatedly
     *** attempting to attach or detach...
     ****************************************************************/
    if(sscanf(friendly_name,"scsi%d",&scsi_bus)==1){
	if(scsi_attach(infile,sizeof(infile),this)==0){
	    int fd = open(infile,O_RDONLY);
	    if(fd>0){
		return fd;
	    }
	    /* attach was successful but open failed. */
	    imaging_failed = true;
	    return FD_IDENT;
	}
    }
    return -1;
}





/*
 * main image loop.
 * if high_water_mark==0, then we do not know how many blocks the input
 * is; just read it byte-by-byte...
 */
void imager::image_loop(uint64 low_water_mark, // sector # to start
			uint64 high_water_mark, // sector # to end
			int direction, int readsectors,int error_mask)
{
    // buffer to store the data we read
    bufsize = readsectors*sector_size;
    buf = (unsigned char *)malloc(bufsize); 
    memset(buf,0,sizeof(buf));
    uint64 data_offset = 0;		// offset into output file

    if(!buf) err(1,"malloc");

    /* Loop as long as we have room, or until we get an EOF
     * (if high_water_mark is 0.)
     */
    imaging = true;
    while(low_water_mark < high_water_mark || high_water_mark==0){ 

	/* Figure out where to read and how how many sectors to read */
	uint64 snum;	// where we will be reading
	unsigned int sectors_to_read = readsectors;
	if(sectors_to_read > maxreadblocks && maxreadblocks>0){
	    sectors_to_read = maxreadblocks;
	}
	if(direction==1){	// going up
	    snum = low_water_mark;

	    /* If a high water mark is set, take it into account */
	    if(high_water_mark>0){
		unsigned int sectors_left = high_water_mark - snum;
		if(sectors_left < sectors_to_read){ 
		    sectors_to_read = sectors_left;
		}
	    }
	}
	else {
	    assert(high_water_mark != 0); // we can't go backwards if we don't know end

	    snum = high_water_mark - sectors_to_read;
	    if(snum<low_water_mark){
		snum = low_water_mark;
		sectors_to_read = high_water_mark - low_water_mark;
	    }
	}

	last_sector_read = snum;
	last_sectors_read = sectors_to_read;
	last_direction = direction;

	data_offset = sector_size * snum; // where we want to start reading
	lseek(in,data_offset,SEEK_SET); // make sure we are at the right place; (ignore error)

	status();			// tell the user what we are doing

	/* Now seek and read */
	read_timer.start();

	int bytes_to_read = sectors_to_read * sector_size;
	int bytes_read    = read(in,buf,bytes_to_read);
	read_timer.stop();

	if(bytes_read == bytes_to_read){
	    /* Got a good read! */
	    total_sectors_read    += sectors_to_read;
	    total_bytes_read      += bytes_read;
	    errors_in_a_row = 0;

	    /* Write the data! */
	    write_data(buf,data_offset,bytes_read);

	    if(direction==1){
		low_water_mark += sectors_to_read;
	    }
	    else {
		high_water_mark -= sectors_to_read;
	    }
	    continue;
	}

	/* If high water mark is 0,
	 * then just write out what we read and continue, because we don't know how many
	 * bytes we can read...
	 */
	if(high_water_mark==0){
	    total_bytes_read      += bytes_read; 
	    if(bytes_read<=0) break;	// end of pipe/file/whatever
	    write_data(buf,data_offset,bytes_read);
	    data_offset += bytes_read;	// move along
	    continue;
	}

	/* Error handling follows */
	hash_invalid = true;
	if(error_mask==0){
	    /* Fill buffer that wasn't read with bad block flag
	     * and write it out.
	     */
	    for(unsigned char *cc=buf; cc < buf+bytes_to_read-AFF_SECTOR_SIZE;
		cc += AFF_SECTOR_SIZE){
		memcpy(cc,af_badflag(af),AFF_SECTOR_SIZE);
	    }
	    write_data(buf,data_offset,bytes_to_read);
	    bad_sectors_read += sectors_to_read;
	    snum += sectors_to_read*direction;	// increment sector number counter
	    errors_in_a_row += 1;

	    /* See if the errors exceed the thresshold */
	    if(errors_in_a_row>retry_count){
		if(direction == -1){
		    /* That's it. Give up */
		    break;
		}
		errors_in_a_row = 0; // reset count
		direction = -1;
		continue;
	    }
	}
	if(error_mask==1){
	    /* Stop reading at the first error and write the incomplete buffer */
	    if(bytes_read>0){
		write_data(buf,data_offset,bytes_read);
	    }
	    break;
	}
    }
    imaging = false;
    free(buf);
}

/* Returns 0 if okay, -1 if failure. */
int imager::set_input_fd(int ifd)
{
    in = ifd;

    /* Make sure infile is actually a device, and not a file */
    struct stat so;
    if(fstat(ifd,&so)){
	perror("fstat");
    }
    int mode = so.st_mode & S_IFMT;
    struct af_figure_media_buf afb;
    memset(&afb,0,sizeof(afb));

    /* Now figure out how many input blocks we have */
    if(mode==S_IFBLK || mode==S_IFCHR){
	if (af_figure_media(in,&afb)){
	    return -1;
	}
	sector_size = afb.sector_size;
	total_sectors = afb.total_sectors;
	maxreadblocks = afb.max_read_blocks;
	return 0;
    }
    if(mode==S_IFREG && allow_regular==false){
	fprintf(stderr,"input is a regular file.\n");
	fprintf(stderr,"Use aconvert to convert regular files to AFF.\n");
	return -1;
    }

    /* Okay. We don't know how big it will be, so just get what we can... */
    sector_size   = 512;		// it's a good guess
    total_sectors = 0;			// we don't know
    maxreadblocks = 0;			// no limit
    return 0;
}


void imager::set_input(const char *name)
{
    /* Set the input given a "name"
     * 
     * First, try to open the input file.
     * If the name specified by the user can be opened, use it.
     * If the name cannot be opened work, see if it is a operating system
     * specific filename such as "ide0" or "ata0", in which the
     * operating system-specific code will attempt to attach the
     * device.
     */
    /* Check for '-' which is stdin */
    if(strcmp(name,"-")==0){
	strcpy(infile,name);		// make a local copy
	set_input_fd(0);			// file descriptor 0 is stdin
	return;
    }

    /* Check for 'listen:%d' which means listen for a TCP connection */
    int port;
    if(sscanf(name,"listen:%d",&port)==1){
	socket_listen(port);		// sets infile
	return;
    }

    /* The name must be a file. See if we can open it... */
    int ifd = open(name,O_RDONLY);
    if(ifd>0){
	strcpy(infile,name);		// make a local copy
	set_input_fd(ifd);
	return;
    }

    /* Attempt to open infile failed; check for a special
     * device name...
     */
    ifd = open_dev(name);
    if(ifd>0){
	set_input_fd(ifd);
	return;
    }

    /* If we haven't been able to open the something by this point, give up. */
    perror(name);
    exit(1);
}



void imager::hash_setup()
{
    /* Set up the MD5 & SHA1 machinery */
    MD5_Init(&md5);
    SHA1_Init(&sha);
}


/* start_imaging2():
 * Actually run the imaging
 */
void imager::start_imaging2()
{
    retry_count = opt_retry_count;

    /* See if the skipping makes sense */
    if(!opt_skip_sectors){
	if(opt_skip % sector_size != 0){
	    fprintf(stderr,
		    "Skipping must be an integral multiple of sector size "
		    "(%d bytes)\n",sector_size);
	    imaging_failed = true;
	    return;
	}
	opt_skip /= sector_size;	// get the actuall offset
    }

    int starting_direction = 1;
    if(opt_reverse) starting_direction = -1;

    /****************************************************************
     *** Start imaging
     ****************************************************************/

    signal(SIGINT,sig_intr);	// set the signal handler
    hash_setup();		// get ready...
    image_loop(opt_skip,
	       total_sectors,starting_direction,
	       opt_readsectors,opt_error_count); // start the process
    signal(SIGINT,0);		// unset the handler
    signal(SIGCONT,sig_cont);

    /****************************************************************
     *** Finished imaging
     ****************************************************************/

    /* Calculate the final MD5 and SHA1 */
    MD5_Final(final_md5,&md5);
    SHA1_Final(final_sha1,&sha);
}




/* Start the imaging.
 * If files are specified, opens them.
 * then does the imaging.
 * Then closes the files.
 */
int imager::start_imaging()
{
    if(opt_zap){
	if(fname_raw[0]) unlink(fname_raw);	// if we are zapping
	if(fname_aff[0]) unlink(fname_aff);
    }

    /* If either file exists and we are not appending, give an error */
    if(!opt_append){
	if(access(fname_raw,F_OK)==0) errx(1,"%s: file exists",fname_raw);
	if(access(fname_aff,F_OK)==0) errx(1,"%s: file exists",fname_aff);
    }

    if(fname_raw[0]){
	fout = fopen(fname_raw,"wb");
	if(!fout){
	  err(1,"fopen(%s)",fname_raw);
	}
    }
	
    /* If the user is imaging to an AFF file,
     * open it and try to ident the drive.
     * Drive ident is not done if writing to a raw file, because
     * there is no place to store the ident information. This will be changed
     * when we can write an XML log.
     */
    if(fname_aff[0]){
	char buf[256];
	memset(buf,0,sizeof(buf));

	if(af_identify_file(fname_aff)==0){
	    fprintf(stderr,"%s exists and is not an AFF file.\n",fname_aff);
	    fprintf(stderr,"Delete it or move it first.\n");
	    exit(-1);
	}

	af = af_open(fname_aff,O_CREAT|O_RDWR,0666);
	if(!af){
	    err(1,"af_open %s: ",fname_aff);
	}
	af->tag = (void *)this;			// remember me!
	if(opt_logAFF){
	    af->logfile = logfile;
	}

	/* If the segment size hasn't been set, then set it */
	af_update_seg(af,AF_MD5,0,buf,16,1);	// leave 16 bytes for MD5
	af_update_seg(af,AF_SHA1,0,buf,20,1);	// leave 20 bytes for SHA1
	af_set_callback(af,segwrite_callback);
	af_set_sizes(af,opt_pagesize,(uint64)0);		// 
	af_enable_writing(af,1);	// we will be writing this file
	af_enable_compression(af,opt_compression_type,opt_compression_level);
	process_config_questions(af,this); // process the config file for questions
	ident();			// ident the drive
	af_update_seg(af,AF_IMAGING_COMMAND_LINE,0,command_line,strlen(command_line),1);
	af_update_seg(af,AF_IMAGING_DEVICE,0,infile,strlen(infile),1);
	unsigned char bit128[16];
	RAND_pseudo_bytes(bit128,sizeof(bit128));
	af_update_seg(af,AF_IMAGE_GID,0,bit128,sizeof(bit128),1);

	time_t t = time(0);
	strftime(buf,sizeof(buf),"%Y-%m-%d %H:%M:%S",localtime(&t));
	af_update_seg(af,AF_IMAGING_DATE,0,buf,strlen(buf),1);
    }

    /* Here is where the imaging takes place.
     * Do it unless ifd==FD_IDENT, which is the fictitious FD.
     */

    if(logfile){
	fprintf(logfile,"aimage infile=%s ",infile);
	if(fname_raw[0]) fprintf(logfile,"outfile_raw=%s ",fname_raw);
	if(fname_aff[0]) fprintf(logfile,"outfile_aff=%s ",fname_aff);
	fprintf(logfile,"\n");
    }
    if(in!=FD_IDENT){
	imaging_timer.start();
	start_imaging2();
	imaging_timer.stop();
    }


    /* Raw Cleanup */
    if(fout){
	if(fclose(fout)){
	    warn("fclose(%s)",fname_raw);
	    fprintf(stderr,"This shouldn't happen. '%s' may be corrupt.\n", fname_raw);
	}
	fout = 0;
    }


    /* AFF Cleanup... */
    if(af){
	if(hash_invalid==false){
	    if(af_update_seg(af,AF_MD5,0,final_md5,16,0)){
		perror("Could not update AF_MD5");
	    }
	    if(af_update_seg(af,AF_SHA1,0,final_sha1,20,0)){
		perror("Could not update AF_SHA1");
	    }
	}
	else {
	    af_del_seg(af,AF_MD5);	// because it is not valid
	    af_del_seg(af,AF_SHA1);
	}
	if(af_update_segq(af,AF_BADSECTORS, bad_sectors_read,1)){
	    perror("Could not update AF_BADSECTORS");
	}
	if(af_update_segq(af,AF_BLANKSECTORS, total_blank_sectors,1)){
	    perror("Could not update AF_BLANKSECTORS");
	}

	if(af_close(af)){
	    perror("af_close");
	    fprintf(stderr,"af_close failed. This shouldn't happen.\n");
	    fprintf(stderr,"Run 'ainfo -v %s' to see if %s is corrupt.\n",
		    fname_aff,fname_aff);
	}
	af = 0;
    }
    return 0;
}



/* Listen for a local socket connection and return the
 * file descriptor...
 */
int imager::socket_listen(int port)
{
    struct sockaddr_in local;
    struct sockaddr_in remote;
    socklen_t rsize = sizeof(remote);

    int sock = socket(AF_INET,SOCK_STREAM,IPPROTO_IP);    /* Open a listening socket ... */
    memset(&local,0,sizeof(local));
    memset(&remote,0,sizeof(remote));
#ifndef linux				// have I mentioned how much I hate linux?
    local.sin_len = sizeof(sockaddr_in);
#endif
    local.sin_family = AF_INET;
    local.sin_port   = htonl(port);	// listen on port 5555
    if(bind(sock,(sockaddr *)&local,sizeof(local))) err(1,"bind");
    if(listen(sock,0)) err(1,"listen");		// listen, and only accept one
    printf("Listening for connection on port %d...\n",port);
    in = accept(sock,(sockaddr *)&remote,&rsize);
    if(in<0){
	perror("accept");
	in = 0;
	return -1;
    }
    strcpy(infile,inet_ntoa(remote.sin_addr));
    printf("Connection accepted from %s\n",infile);
    return 0;
}



/* final_report():
 * Let's make the user feel good...
 */
void imager::final_report()
{
    bold("****************************** IMAGING REPORT ******************************");
    putchar('\n');
    printf("Input: "); bold(infile); putchar('\n');
    if(device_model[0]){
	printf("  Model: ");
	bold(device_model);
    }
    if(serial_number[0]){
	printf("  S/N: ");
	bold(serial_number);
    }
    putchar('\n');
    if(fname_aff[0]){
	printf("  AFF Output file: ");
	bold(fname_aff);
    }
    if(fname_raw[0]){
	printf("   Raw Output file: ");
	bold(fname_raw);
    }

    putchar('\n');
    
    printf("  Bytes read: ");
    comma_printf(total_bytes_read,0);
    printf("    Bytes written:");
    comma_printf(callback_bytes_written,0);
    putchar('\n');
    
    char print_buf[256];
    printf("\n");
    if(hash_invalid==false){
	printf("raw image md5:  %s\n",
	       af_hexbuf(print_buf,sizeof(print_buf),final_md5,16,1));
	
	printf("raw image sha1: %s\n",
	       af_hexbuf(print_buf,sizeof(print_buf),final_sha1,20,1));
    }

    if(imaging_failed){
	printf("\nTHIS DRIVE COULD NOT BE IMAGED DUE TO A HARDWARE FAILURE.\n");
    }
}

