#include "afflib.h"
#include "afflib_i.h"


/* evf file implementation */
struct evf_private {
    /* private data */
    int something;
};

static inline struct evf_private *EVF_PRIVATE(AFFILE *af)
{
    return (struct evf_private *)(af->vnodeprivate);
}

/* Return 1 if a file is a evf file... */
static int evf_identify_file(const char *filename)
{
    return access(filename,R_OK)==0;	// if we can read it, it's evf...
}


/* Return the size of the evf file */
static int64 evf_filesize(AFFILE *af)
{
    struct stat sb;
    if(fstat(fileno(af->evf),&sb)){
	warn("evf_open: stat(%s): ",af->fname);
	return -1;			// kind of odd...
    }
    return sb.st_size;
}

static int evf_open(AFFILE *af)
{
    /* Evf is the passthrough system.
     * Right now, it is read only...
     */
    af->evf=fopen(af->fname,"rb");
    af->imagesize = evf_filesize(af);
    af->image_pagesize = EVF_PAGESIZE;
    af->vnodeprivate = (void *)calloc(1,sizeof(struct evf_private));
    struct evf_private *rp = EVF_PRIVATE(af);
    rp->cur_page = 0;

    return 0;
}

static int evf_close(AFFILE *af)
{
    struct evf_private *rp = EVF_PRIVATE(af);

    if(af->evf_popen){
	pclose(af->evf);
    }
    else {
	fclose(af->evf);
    }
    memset(rp,0,sizeof(*rp));		// clean object reuse
    free(rp);				// won't need it again
    return 0;
}

static int evf_get_seg(AFFILE *af,const char *name,unsigned long *arg,unsigned char *data,size_t *datalen)
{
    /* Right now, this only supports page segments */
    unsigned int segnum=0;
    if(sscanf(name,AF_PAGE,&segnum)!=1){
	return -1;			// unknown segment number
    }
    fflush(af->evf);			// make sure that any buffers are flushed

    int64 pos = (int64)segnum * af->image_pagesize; // where we are to start reading
    int64 bytes_left = af->imagesize - pos;	// how many bytes left in the file

    int bytes_to_read = af->image_pagesize; // copy this many bytes, unless
    if(bytes_to_read > bytes_left) bytes_to_read = bytes_left; // only this much is left
    
    if(arg) *arg = 0;			// arg is always 0
    if(datalen){
	if(*datalen==0 && data==0){ // asked for 0 bytes, so give the actual size
	    *datalen = bytes_to_read;
	    return 0;
	}
	if(*datalen < (unsigned)bytes_to_read){
	    *datalen = bytes_to_read;
	    return AF_ERROR_DATASMALL;
	}
    }
    if(data){
	fseeko(af->evf,pos,SEEK_SET);    
	int bytes_read = fread(data,1,bytes_to_read,af->evf);
	if(bytes_read==bytes_to_read) return 0;
	return -1;			// some kind of EOF?
    }
    return 0;				// no problems!
}


int evf_update_seg(AFFILE *af, const char *name,
		    unsigned long arg,const void *value,unsigned int vallen,int append)
{
    /* Simple implementation; only updates data segments */
    unsigned int segnum=0;
    if(sscanf(name,AF_PAGE,&segnum)!=1){
	return -1;			// unknown segment number
    }
    int64 pos = (int64)segnum * af->image_pagesize; // where we are to start reading
    fseeko(af->evf,pos,SEEK_SET);
    if(fwrite(value,1,vallen,af->evf)==vallen){
	return 0;
    }
    return -1;				// some kind of error...
}


static int evf_stat(AFFILE *af,struct af_vnode_info *vni)
{
    vni->imagesize = -1;
    vni->pagesize  = EVF_PAGESIZE;		// decent page size

    /* If we can stat the file, use that. */
    fflush(af->evf);
    struct stat sb;
    if(fstat(fileno(af->evf),&sb)==0){
	if(sb.st_mode & S_IFREG){	// only do this for regular files
	    vni->imagesize = sb.st_size;
	}
    }

    if(vni->imagesize==-1){
	/* See if this is a device that we can figure */
	struct af_figure_media_buf afb;
	if(af_figure_media(fileno(af->evf),&afb)==0){
	    if(afb.total_sectors>0 && afb.sector_size>0){
		vni->imagesize = afb.total_sectors * afb.sector_size;
	    }
	}
    }
    return 0;
}

static int evf_rewind_seg(AFFILE *af)
{
    struct evf_private *rp = EVF_PRIVATE(af);
    
    rp->cur_page = 0;
    return 0;
}


static int evf_get_next_seg(AFFILE *af,char *segname,size_t segname_len,unsigned long *arg,
			unsigned char *data,size_t *datalen_)
{
    struct evf_private *rp = EVF_PRIVATE(af);
    char pagename[AF_MAX_NAME_LEN];		//
    
    /* See if we are at the end of the "virtual" segment list */
    if((rp->cur_page)*af->image_pagesize > af->imagesize) return -1;

    /* Make the segment name */
    memset(pagename,0,sizeof(pagename));
    snprintf(pagename,sizeof(pagename),AF_PAGE,rp->cur_page++);

    /* Get the segment, if we can */
    int r = evf_get_seg(af,pagename,arg,data,datalen_);

    /* If r==0 and there is room for copying in the segment name, return it */
    if(r==0){
	if(strlen(pagename)+1 < segname_len){
	    strcpy(segname,pagename);
	    return 0;
	}
	/* segname wasn't big enough */
	return -2;
    }
    return r;				// some other error
}


struct af_vnode vnode_evf = {
    AF_IDENTIFY_EVF,
    evf_identify_file,
    evf_open,
    evf_close,
    evf_stat,
    evf_get_seg,			// get seg
    evf_get_next_seg,			// get_next_seg
    evf_rewind_seg,			// rewind_seg
    evf_update_seg,			// update_seg
    0					// del_seg
};

